@@ -46,8 +46,8 @@
 #include "http_config.h"
 #include "http_protocol.h"
 #include "http_core.h"
-#include "http_log.h"		/* for errors in parse_htaccess */
-#include "http_request.h"	/* for default_handler (see invoke_handler) */
+#include "http_log.h"      /* for errors in parse_htaccess */
+#include "http_request.h"  /* for default_handler (see invoke_handler) */
 #include "http_main.h"
 #include "http_vhost.h"
 #include "util_cfgtree.h"
@@ -73,6 +73,7 @@ APR_HOOK_STRUCT(
            APR_HOOK_LINK(handler)
            APR_HOOK_LINK(quick_handler)
            APR_HOOK_LINK(optional_fn_retrieve)
+           APR_HOOK_LINK(test_config)
 )
 
 AP_IMPLEMENT_HOOK_RUN_ALL(int, header_parser,
@@ -83,6 +84,10 @@ AP_IMPLEMENT_HOOK_RUN_ALL(int, pre_config,
                            apr_pool_t *ptemp),
                           (pconf, plog, ptemp), OK, DECLINED)
 
+AP_IMPLEMENT_HOOK_VOID(test_config,
+                       (apr_pool_t *pconf, server_rec *s),
+                       (pconf, s))
+
 AP_IMPLEMENT_HOOK_RUN_ALL(int, post_config,
                           (apr_pool_t *pconf, apr_pool_t *plog,
                            apr_pool_t *ptemp, server_rec *s),
@@ -177,6 +182,8 @@ static int dynamic_modules = 0;
 AP_DECLARE_DATA module *ap_top_module = NULL;
 AP_DECLARE_DATA module **ap_loaded_modules=NULL;
 
+static apr_hash_t *ap_config_hash = NULL;
+
 typedef int (*handler_func)(request_rec *);
 typedef void *(*dir_maker_func)(apr_pool_t *, char *);
 typedef void *(*merger_func)(apr_pool_t *, void *, void *);
@@ -309,7 +316,7 @@ static int ap_invoke_filter_init(ap_filter_t *filters)
             }
         }
         filters = filters->next;
-    } 
+    }
     return OK;
 }
 
@@ -319,6 +326,7 @@ AP_CORE_DECLARE(int) ap_invoke_handler(request_rec *r)
     const char *p;
     int result;
     const char *old_handler = r->handler;
+    const char *ignore;
 
     /*
      * The new insert_filter stage makes the most sense here.  We only use
@@ -369,6 +377,23 @@ AP_CORE_DECLARE(int) ap_invoke_handler(request_rec *r)
         ap_log_rerror(APLOG_MARK, APLOG_WARNING, 0, r,
             "handler \"%s\" not found for: %s", r->handler, r->filename);
     }
+    if ((result != OK) && (result != DONE) && (result != DECLINED)
+        && (result != AP_FILTER_ERROR)
+        && !ap_is_HTTP_VALID_RESPONSE(result)) {
+        /* If a module is deliberately returning something else
+         * (request_rec in non-HTTP or proprietary extension?)
+         * let it set a note to allow it explicitly.
+         * Otherwise, a return code that is neither reserved nor HTTP
+         * is a bug, as in PR#31759.
+         */
+        ignore = apr_table_get(r->notes, "HTTP_IGNORE_RANGE");
+        if (!ignore) {
+            ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
+                          "Handler for %s returned invalid result code %d",
+                          r->handler, result);
+            result = HTTP_INTERNAL_SERVER_ERROR;
+        }
+    }
 
     return result == DECLINED ? HTTP_INTERNAL_SERVER_ERROR : result;
 }
@@ -403,24 +428,80 @@ AP_DECLARE(void) ap_register_hooks(module *m, apr_pool_t *p)
     }
 }
 
+static void ap_add_module_commands(module *m, apr_pool_t *p);
+
+typedef struct ap_mod_list_struct ap_mod_list;
+struct ap_mod_list_struct {
+    struct ap_mod_list_struct *next;
+    module *m;
+    const command_rec *cmd;
+};
+
+static apr_status_t reload_conf_hash(void *baton)
+{
+    ap_config_hash = NULL;
+    return APR_SUCCESS;
+}
+
+static void rebuild_conf_hash(apr_pool_t *p, int add_prelinked)
+{
+    module **m;
+
+    ap_config_hash = apr_hash_make(p);
+
+    apr_pool_cleanup_register(p, NULL, reload_conf_hash,
+                              apr_pool_cleanup_null);
+    if (add_prelinked) {
+        for (m = ap_prelinked_modules; *m != NULL; m++) {
+            ap_add_module_commands(*m, p);
+        }
+    }
+}
+
+static void ap_add_module_commands(module *m, apr_pool_t *p)
+{
+    apr_pool_t *tpool;
+    ap_mod_list *mln;
+    const command_rec *cmd;
+    char *dir;
+
+    cmd = m->cmds;
+
+    if (ap_config_hash == NULL) {
+        rebuild_conf_hash(p, 0);
+    }
+
+    tpool = apr_hash_pool_get(ap_config_hash);
+
+    while (cmd && cmd->name) {
+        mln = apr_palloc(tpool, sizeof(ap_mod_list));
+        mln->cmd = cmd;
+        mln->m = m;
+        dir = apr_pstrdup(tpool, cmd->name);
+
+        ap_str_tolower(dir);
+
+        mln->next = apr_hash_get(ap_config_hash, dir, APR_HASH_KEY_STRING);
+        apr_hash_set(ap_config_hash, dir, APR_HASH_KEY_STRING, mln);
+        ++cmd;
+    }
+}
+
+
 /* One-time setup for precompiled modules --- NOT to be done on restart */
 
-AP_DECLARE(void) ap_add_module(module *m, apr_pool_t *p)
+AP_DECLARE(const char *) ap_add_module(module *m, apr_pool_t *p)
 {
-    /* This could be called from an AddModule httpd.conf command,
+    /* This could be called from a LoadModule httpd.conf command,
      * after the file has been linked and the module structure within it
      * teased out...
      */
 
     if (m->version != MODULE_MAGIC_NUMBER_MAJOR) {
-        ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, NULL,
-                     "%s: module \"%s\" is not compatible with this "
-                     "version of Apache (found %d, need %d).",
-                     ap_server_argv0, m->name, m->version,
-                     MODULE_MAGIC_NUMBER_MAJOR);
-        ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, NULL,
-                     "Please contact the vendor for the correct version.");
-        exit(1);
+        return apr_psprintf(p, "Module \"%s\" is not compatible with this "
+                            "version of Apache (found %d, need %d). Please "
+                            "contact the vendor for the correct version.",
+                            m->name, m->version, MODULE_MAGIC_NUMBER_MAJOR);
     }
 
     if (m->next == NULL) {
@@ -433,13 +514,10 @@ AP_DECLARE(void) ap_add_module(module *m, apr_pool_t *p)
         dynamic_modules++;
 
         if (dynamic_modules > DYNAMIC_MODULE_LIMIT) {
-            ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, NULL,
-                         "%s: module \"%s\" could not be loaded, because"
-                         " the dynamic", ap_server_argv0, m->name);
-            ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, NULL,
-                         "module limit was reached. Please increase "
-                         "DYNAMIC_MODULE_LIMIT and recompile.");
-            exit(1);
+            return apr_psprintf(p, "Module \"%s\" could not be loaded, "
+                                "because the dynamic module limit was "
+                                "reached. Please increase "
+                                "DYNAMIC_MODULE_LIMIT and recompile.", m->name);
         }
     }
 
@@ -467,10 +545,13 @@ AP_DECLARE(void) ap_add_module(module *m, apr_pool_t *p)
     }
 #endif /*_OSD_POSIX*/
 
+    ap_add_module_commands(m, p);
     /*  FIXME: is this the right place to call this?
      *  It doesn't appear to be
      */
     ap_register_hooks(m, p);
+
+    return NULL;
 }
 
 /*
@@ -518,14 +599,18 @@ AP_DECLARE(void) ap_remove_module(module *m)
     total_modules--;
 }
 
-AP_DECLARE(void) ap_add_loaded_module(module *mod, apr_pool_t *p)
+AP_DECLARE(const char *) ap_add_loaded_module(module *mod, apr_pool_t *p)
 {
     module **m;
+    const char *error;
 
     /*
      *  Add module pointer to top of chained module list
      */
-    ap_add_module(mod, p);
+    error = ap_add_module(mod, p);
+    if (error) {
+        return error;
+    }
 
     /*
      *  And module pointer to list of loaded modules
@@ -539,6 +624,8 @@ AP_DECLARE(void) ap_add_loaded_module(module *mod, apr_pool_t *p)
         ;
     *m++ = mod;
     *m = NULL;
+
+    return NULL;
 }
 
 AP_DECLARE(void) ap_remove_loaded_module(module *mod)
@@ -571,13 +658,16 @@ AP_DECLARE(void) ap_remove_loaded_module(module *mod)
     *m = NULL;
 }
 
-AP_DECLARE(void) ap_setup_prelinked_modules(process_rec *process)
+AP_DECLARE(const char *) ap_setup_prelinked_modules(process_rec *process)
 {
     module **m;
     module **m2;
+    const char *error;
 
     apr_hook_global_pool=process->pconf;
 
+    rebuild_conf_hash(process->pconf, 0);
+
     /*
      *  Initialise total_modules variable and module indices
      */
@@ -592,8 +682,7 @@ AP_DECLARE(void) ap_setup_prelinked_modules(process_rec *process)
         sizeof(module *) * (total_modules + DYNAMIC_MODULE_LIMIT + 1));
 
     if (ap_loaded_modules == NULL) {
-        ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, NULL,
-                     "Ouch!  Out of memory in ap_setup_prelinked_modules()!");
+        return "Ouch! Out of memory in ap_setup_prelinked_modules()!";
     }
 
     for (m = ap_preloaded_modules, m2 = ap_loaded_modules; *m != NULL; )
@@ -604,10 +693,16 @@ AP_DECLARE(void) ap_setup_prelinked_modules(process_rec *process)
     /*
      *   Initialize chain of linked (=activate) modules
      */
-    for (m = ap_prelinked_modules; *m != NULL; m++)
-        ap_add_module(*m, process->pconf);
+    for (m = ap_prelinked_modules; *m != NULL; m++) {
+        error = ap_add_module(*m, process->pconf);
+        if (error) {
+            return error;
+        }
+    }
 
     apr_hook_sort_all();
+
+    return NULL;
 }
 
 AP_DECLARE(const char *) ap_find_module_name(module *m)
@@ -627,26 +722,6 @@ AP_DECLARE(module *) ap_find_linked_module(const char *name)
     return NULL;
 }
 
-/* Add a named module.  Returns 1 if module found, 0 otherwise.  */
-AP_DECLARE(int) ap_add_named_module(const char *name, apr_pool_t *p)
-{
-    module *modp;
-    int i = 0;
-
-    for (modp = ap_loaded_modules[i]; modp; modp = ap_loaded_modules[++i]) {
-        if (strcmp(modp->name, name) == 0) {
-            /* Only add modules that are not already enabled.  */
-            if (modp->next == NULL) {
-                ap_add_module(modp, p);
-            }
-
-            return 1;
-        }
-    }
-
-    return 0;
-}
-
 /*****************************************************************
  *
  * Resource, access, and .htaccess config files now parsed by a common
@@ -656,6 +731,8 @@ AP_DECLARE(int) ap_add_named_module(const char *name, apr_pool_t *p)
  * invoking the function...
  */
 
+#define AP_MAX_ARGC 64
+
 static const char *invoke_cmd(const command_rec *cmd, cmd_parms *parms,
                               void *mconfig, const char *args)
 {
@@ -675,6 +752,23 @@ static const char *invoke_cmd(const command_rec *cmd, cmd_parms *parms,
 #endif
         return cmd->AP_RAW_ARGS(parms, mconfig, args);
 
+    case TAKE_ARGV:
+        {
+            char *argv[AP_MAX_ARGC];
+            int argc = 0;
+
+            do {
+                w = ap_getword_conf(parms->pool, &args);
+                if (*w == '\0' && *args == '\0') {
+                    break;
+                }
+                argv[argc] = w;
+                argc++;
+            } while (argc < AP_MAX_ARGC && *args != '\0');
+
+            return cmd->AP_TAKE_ARGV(parms, mconfig, argc, argv);
+        }
+
     case NO_ARGS:
         if (*args != 0)
             return apr_pstrcat(parms->pool, cmd->name, " takes no arguments",
@@ -997,7 +1091,7 @@ AP_DECLARE(const char *) ap_build_cont_config(apr_pool_t *p,
     ap_directive_t *sub_tree = NULL;
 
     /* Since this function can be called recursively, allocate
-     * the temporary 8k string buffer from the temp_pool rather 
+     * the temporary 8k string buffer from the temp_pool rather
      * than the stack to avoid over-running a fixed length stack.
      */
     l = apr_palloc(temp_pool, MAX_STRING_LEN);
@@ -1031,54 +1125,53 @@ static const char *ap_walk_config_sub(const ap_directive_t *current,
                                       cmd_parms *parms,
                                       ap_conf_vector_t *section_vector)
 {
-    module *mod = ap_top_module;
+    const command_rec *cmd;
+    ap_mod_list *ml;
+    char *dir = apr_pstrdup(parms->pool, current->directive);
 
-    while (1) {
-        const command_rec *cmd;
+    ap_str_tolower(dir);
 
-        if (!(cmd = ap_find_command_in_modules(current->directive, &mod))) {
-            parms->err_directive = current;
-            return apr_pstrcat(parms->pool, "Invalid command '",
-                               current->directive,
-                               "', perhaps mis-spelled or defined by a module "
-                               "not included in the server configuration",
-                               NULL);
+    ml = apr_hash_get(ap_config_hash, dir, APR_HASH_KEY_STRING);
+
+    if (ml == NULL) {
+        parms->err_directive = current;
+        return apr_pstrcat(parms->pool, "Invalid command '",
+                           current->directive,
+                           "', perhaps misspelled or defined by a module "
+                           "not included in the server configuration",
+                           NULL);
+    }
+
+    for ( ; ml != NULL; ml = ml->next) {
+        void *dir_config = ap_set_config_vectors(parms->server,
+                                                 section_vector,
+                                                 parms->path,
+                                                 ml->m,
+                                                 parms->pool);
+        const char *retval;
+        cmd = ml->cmd;
+
+        /* Once was enough? */
+        if (cmd->req_override & EXEC_ON_READ) {
+            continue;
         }
-        else {
-            void *dir_config = ap_set_config_vectors(parms->server,
-                                                     section_vector,
-                                                     parms->path,
-                                                     mod,
-                                                     parms->pool);
-            const char *retval;
 
-            /* Once was enough? */
-            if (cmd->req_override & EXEC_ON_READ) {
-                return NULL;
+        retval = invoke_cmd(cmd, parms, dir_config, current->args);
+
+        if (retval != NULL && strcmp(retval, DECLINE_CMD) != 0) {
+            /* If the directive in error has already been set, don't
+             * replace it.  Otherwise, an error inside a container
+             * will be reported as occuring on the first line of the
+             * container.
+             */
+            if (!parms->err_directive) {
+                parms->err_directive = current;
             }
-
-            retval = invoke_cmd(cmd, parms, dir_config, current->args);
-            if (retval == NULL) {
-                return NULL;
-            }
-
-            if (strcmp(retval, DECLINE_CMD) != 0) {
-                /* If the directive in error has already been set, don't
-                 * replace it.  Otherwise, an error inside a container 
-                 * will be reported as occuring on the first line of the
-                 * container.
-                 */
-                if (!parms->err_directive) {
-                    parms->err_directive = current;
-                }
-
-                return retval;
-            }
-
-            mod = mod->next; /* Next time around, skip this one */
+            return retval;
         }
     }
-    /* NOTREACHED */
+
+    return NULL;
 }
 
 AP_DECLARE(const char *) ap_walk_config(ap_directive_t *current,
@@ -1260,7 +1353,7 @@ AP_DECLARE(char *) ap_server_root_relative(apr_pool_t *p, const char *file)
     apr_status_t rv;
     rv = apr_filepath_merge(&newpath, ap_server_root, file,
                             APR_FILEPATH_TRUENAME, p);
-    if (newpath && (rv == APR_SUCCESS || APR_STATUS_IS_EPATHWILD(rv) 
+    if (newpath && (rv == APR_SUCCESS || APR_STATUS_IS_EPATHWILD(rv)
                                       || APR_STATUS_IS_ENOENT(rv)
                                       || APR_STATUS_IS_ENOTDIR(rv))) {
         return newpath;
@@ -1321,19 +1414,34 @@ static const char *execute_now(char *cmd_line, const char *args,
                                ap_directive_t **sub_tree,
                                ap_directive_t *parent)
 {
-    module *mod = ap_top_module;
     const command_rec *cmd;
+    ap_mod_list *ml;
+    char *dir = apr_pstrdup(parms->pool, cmd_line);
 
-    if (!(cmd = ap_find_command_in_modules(cmd_line, &mod))) {
+    ap_str_tolower(dir);
+
+    ml = apr_hash_get(ap_config_hash, dir, APR_HASH_KEY_STRING);
+
+    if (ml == NULL) {
         return apr_pstrcat(parms->pool, "Invalid command '",
                            cmd_line,
-                           "', perhaps mis-spelled or defined by a module "
+                           "', perhaps misspelled or defined by a module "
                            "not included in the server configuration",
                            NULL);
     }
-    else {
-        return invoke_cmd(cmd, parms, sub_tree, args);
+
+    for ( ; ml != NULL; ml = ml->next) {
+        const char *retval;
+        cmd = ml->cmd;
+
+        retval = invoke_cmd(cmd, parms, sub_tree, args);
+
+        if (retval != NULL) {
+            return retval;
+        }
     }
+
+    return NULL;
 }
 
 /* This structure and the following functions are needed for the
@@ -1380,9 +1488,11 @@ static int arr_elts_close(void *param)
     return 0;
 }
 
-static void process_command_config(server_rec *s, apr_array_header_t *arr,
-                                   ap_directive_t **conftree, apr_pool_t *p,
-                                   apr_pool_t *ptemp)
+static const char *process_command_config(server_rec *s,
+                                          apr_array_header_t *arr,
+                                          ap_directive_t **conftree,
+                                          apr_pool_t *p,
+                                          apr_pool_t *ptemp)
 {
     const char *errmsg;
     cmd_parms parms;
@@ -1391,26 +1501,30 @@ static void process_command_config(server_rec *s, apr_array_header_t *arr,
     arr_parms.curr_idx = 0;
     arr_parms.array = arr;
 
+    if (ap_config_hash == NULL) {
+        rebuild_conf_hash(s->process->pconf, 1);
+    }
+
     parms = default_parms;
     parms.pool = p;
     parms.temp_pool = ptemp;
     parms.server = s;
     parms.override = (RSRC_CONF | OR_ALL) & ~(OR_AUTHCFG | OR_LIMIT);
+    parms.override_opts = OPT_ALL | OPT_INCNOEXEC | OPT_SYM_OWNER | OPT_MULTI;
 
     parms.config_file = ap_pcfg_open_custom(p, "-c/-C directives",
                                             &arr_parms, NULL,
                                             arr_elts_getstr, arr_elts_close);
 
     errmsg = ap_build_config(&parms, p, ptemp, conftree);
+    ap_cfg_closefile(parms.config_file);
+
     if (errmsg) {
-        ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, NULL,
-                     "Syntax error in -C/-c directive:");
-        ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, NULL,
-                     "%s", errmsg);
-        exit(1);
+        return apr_pstrcat(p, "Syntax error in -C/-c directive: ", errmsg,
+                           NULL);
     }
 
-    ap_cfg_closefile(parms.config_file);
+    return NULL;
 }
 
 typedef struct {
@@ -1425,15 +1539,17 @@ static int fname_alphasort(const void *fn1, const void *fn2)
     return strcmp(f1->fname,f2->fname);
 }
 
-static void process_resource_config_nofnmatch(server_rec *s, const char *fname,
-                                              ap_directive_t **conftree,
-                                              apr_pool_t *p,
-                                              apr_pool_t *ptemp,
-                                              unsigned depth)
+static const char *process_resource_config_nofnmatch(server_rec *s,
+                                                     const char *fname,
+                                                     ap_directive_t **conftree,
+                                                     apr_pool_t *p,
+                                                     apr_pool_t *ptemp,
+                                                     unsigned depth)
 {
     cmd_parms parms;
     ap_configfile_t *cfp;
-    const char *errmsg;
+    const char *error;
+    apr_status_t rv;
 
     if (ap_is_directory(p, fname)) {
         apr_dir_t *dirp;
@@ -1441,15 +1557,13 @@ static void process_resource_config_nofnmatch(server_rec *s, const char *fname,
         int current;
         apr_array_header_t *candidates = NULL;
         fnames *fnew;
-        apr_status_t rv;
-        char errmsg[120], *path = apr_pstrdup(p, fname);
+        char *path = apr_pstrdup(p, fname);
 
         if (++depth > AP_MAX_INCLUDE_DIR_DEPTH) {
-            fprintf(stderr, "%s: Directory %s exceeds the maximum include "
-                    "directory nesting level of %u. You have probably a "
-                    "recursion somewhere.\n", ap_server_argv0, path,
-                    AP_MAX_INCLUDE_DIR_DEPTH);
-            exit(1);
+            return apr_psprintf(p, "Directory %s exceeds the maximum include "
+                                "directory nesting level of %u. You have "
+                                "probably a recursion somewhere.", path,
+                                AP_MAX_INCLUDE_DIR_DEPTH);
         }
 
         /*
@@ -1459,10 +1573,9 @@ static void process_resource_config_nofnmatch(server_rec *s, const char *fname,
          */
         rv = apr_dir_open(&dirp, path, p);
         if (rv != APR_SUCCESS) {
-            fprintf(stderr, "%s: could not open config directory %s: %s\n",
-                    ap_server_argv0, path,
-                    apr_strerror(rv, errmsg, sizeof errmsg));
-            exit(1);
+            char errmsg[120];
+            return apr_psprintf(p, "Could not open config directory %s: %s",
+                                path, apr_strerror(rv, errmsg, sizeof errmsg));
         }
 
         candidates = apr_array_make(p, 1, sizeof(fnames));
@@ -1486,12 +1599,16 @@ static void process_resource_config_nofnmatch(server_rec *s, const char *fname,
              */
             for (current = 0; current < candidates->nelts; ++current) {
                 fnew = &((fnames *) candidates->elts)[current];
-                process_resource_config_nofnmatch(s, fnew->fname, conftree, p,
-                                                  ptemp, depth);
+                error = process_resource_config_nofnmatch(s, fnew->fname,
+                                                          conftree, p, ptemp,
+                                                          depth);
+                if (error) {
+                    return error;
+                }
             }
         }
 
-        return;
+        return NULL;
     }
 
     /* GCC's initialization extensions are soooo nice here... */
@@ -1500,37 +1617,33 @@ static void process_resource_config_nofnmatch(server_rec *s, const char *fname,
     parms.temp_pool = ptemp;
     parms.server = s;
     parms.override = (RSRC_CONF | OR_ALL) & ~(OR_AUTHCFG | OR_LIMIT);
+    parms.override_opts = OPT_ALL | OPT_INCNOEXEC | OPT_SYM_OWNER | OPT_MULTI;
 
-    if (ap_pcfg_openfile(&cfp, p, fname) != APR_SUCCESS) {
-        ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, NULL,
-                     "%s: could not open document config file %s",
-                     ap_server_argv0, fname);
-        exit(1);
+    rv = ap_pcfg_openfile(&cfp, p, fname);
+    if (rv != APR_SUCCESS) {
+        char errmsg[120];
+        return apr_psprintf(p, "Could not open configuration file %s: %s",
+                            fname, apr_strerror(rv, errmsg, sizeof errmsg));
     }
 
     parms.config_file = cfp;
-
-    errmsg = ap_build_config(&parms, p, ptemp, conftree);
-
-    if (errmsg != NULL) {
-        ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, NULL,
-                     "Syntax error on line %d of %s:",
-                     parms.err_directive->line_num,
-                     parms.err_directive->filename);
-        ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, NULL,
-                     "%s", errmsg);
-        exit(1);
-    }
-
+    error = ap_build_config(&parms, p, ptemp, conftree);
     ap_cfg_closefile(cfp);
 
-    return;
+    if (error) {
+        return apr_psprintf(p, "Syntax error on line %d of %s: %s",
+                            parms.err_directive->line_num,
+                            parms.err_directive->filename, error);
+    }
+
+    return NULL;
 }
 
-AP_DECLARE(void) ap_process_resource_config(server_rec *s, const char *fname,
-                                            ap_directive_t **conftree,
-                                            apr_pool_t *p,
-                                            apr_pool_t *ptemp)
+AP_DECLARE(const char *) ap_process_resource_config(server_rec *s,
+                                                    const char *fname,
+                                                    ap_directive_t **conftree,
+                                                    apr_pool_t *p,
+                                                    apr_pool_t *ptemp)
 {
     /* XXX: lstat() won't work on the wildcard pattern...
      */
@@ -1541,12 +1654,13 @@ AP_DECLARE(void) ap_process_resource_config(server_rec *s, const char *fname,
         && !(strcmp(fname, ap_server_root_relative(p, SERVER_CONFIG_FILE)))) {
         apr_finfo_t finfo;
 
-        if (apr_lstat(&finfo, fname, APR_FINFO_TYPE, p) != APR_SUCCESS)
-            return;
+        if (apr_stat(&finfo, fname, APR_FINFO_LINK | APR_FINFO_TYPE, p) != APR_SUCCESS)
+            return NULL;
     }
 
     if (!apr_fnmatch_test(fname)) {
-        process_resource_config_nofnmatch(s, fname, conftree, p, ptemp, 0);
+        return process_resource_config_nofnmatch(s, fname, conftree, p, ptemp,
+                                                 0);
     }
     else {
         apr_dir_t *dirp;
@@ -1555,7 +1669,7 @@ AP_DECLARE(void) ap_process_resource_config(server_rec *s, const char *fname,
         apr_array_header_t *candidates = NULL;
         fnames *fnew;
         apr_status_t rv;
-        char errmsg[120], *path = apr_pstrdup(p, fname), *pattern = NULL;
+        char *path = apr_pstrdup(p, fname), *pattern = NULL;
 
         pattern = ap_strrchr(path, '/');
 
@@ -1564,21 +1678,18 @@ AP_DECLARE(void) ap_process_resource_config(server_rec *s, const char *fname,
         *pattern++ = '\0';
 
         if (apr_fnmatch_test(path)) {
-            fprintf(stderr, "%s: wildcard patterns not allowed in Include "
-                    "%s\n", ap_server_argv0, fname);
-            exit(1);
+            return apr_pstrcat(p, "Wildcard patterns not allowed in Include ",
+                               fname, NULL);
         }
 
-        if (!ap_is_directory(p, path)){ 
-            fprintf(stderr, "%s: Include directory '%s' not found",
-                    ap_server_argv0, path);
-            exit(1);
+        if (!ap_is_directory(p, path)){
+            return apr_pstrcat(p, "Include directory '", path, "' not found",
+                               NULL);
         }
 
         if (!apr_fnmatch_test(pattern)) {
-            fprintf(stderr, "%s: must include a wildcard pattern "
-                    "for Include %s\n", ap_server_argv0, fname);
-            exit(1);
+            return apr_pstrcat(p, "Must include a wildcard pattern for "
+                               "Include ", fname, NULL);
         }
 
         /*
@@ -1588,10 +1699,9 @@ AP_DECLARE(void) ap_process_resource_config(server_rec *s, const char *fname,
          */
         rv = apr_dir_open(&dirp, path, p);
         if (rv != APR_SUCCESS) {
-            fprintf(stderr, "%s: could not open config directory %s: %s\n",
-                    ap_server_argv0, path,
-                    apr_strerror(rv, errmsg, sizeof errmsg));
-            exit(1);
+            char errmsg[120];
+            return apr_psprintf(p, "Could not open config directory %s: %s",
+                                path, apr_strerror(rv, errmsg, sizeof errmsg));
         }
 
         candidates = apr_array_make(p, 1, sizeof(fnames));
@@ -1600,7 +1710,7 @@ AP_DECLARE(void) ap_process_resource_config(server_rec *s, const char *fname,
             if (strcmp(dirent.name, ".")
                 && strcmp(dirent.name, "..")
                 && (apr_fnmatch(pattern, dirent.name,
-                                FNM_PERIOD) == APR_SUCCESS)) {
+                                APR_FNM_PERIOD) == APR_SUCCESS)) {
                 fnew = (fnames *) apr_array_push(candidates);
                 fnew->fname = ap_make_full_path(p, path, dirent.name);
             }
@@ -1608,6 +1718,8 @@ AP_DECLARE(void) ap_process_resource_config(server_rec *s, const char *fname,
 
         apr_dir_close(dirp);
         if (candidates->nelts != 0) {
+            const char *error;
+
             qsort((void *) candidates->elts, candidates->nelts,
                   sizeof(fnames), fname_alphasort);
 
@@ -1617,18 +1729,23 @@ AP_DECLARE(void) ap_process_resource_config(server_rec *s, const char *fname,
              */
             for (current = 0; current < candidates->nelts; ++current) {
                 fnew = &((fnames *) candidates->elts)[current];
-                process_resource_config_nofnmatch(s, fnew->fname, conftree, p,
-                                                  ptemp, 0);
+                error = process_resource_config_nofnmatch(s, fnew->fname,
+                                                          conftree, p,
+                                                          ptemp, 0);
+                if (error) {
+                    return error;
+                }
             }
         }
     }
 
-    return;
+    return NULL;
 }
 
-AP_DECLARE(void) ap_process_config_tree(server_rec *s,
-                                        ap_directive_t *conftree,
-                                        apr_pool_t *p, apr_pool_t *ptemp)
+AP_DECLARE(int) ap_process_config_tree(server_rec *s,
+                                       ap_directive_t *conftree,
+                                       apr_pool_t *p,
+                                       apr_pool_t *ptemp)
 {
     const char *errmsg;
     cmd_parms parms;
@@ -1638,6 +1755,7 @@ AP_DECLARE(void) ap_process_config_tree(server_rec *s,
     parms.temp_pool = ptemp;
     parms.server = s;
     parms.override = (RSRC_CONF | OR_ALL) & ~(OR_AUTHCFG | OR_LIMIT);
+    parms.override_opts = OPT_ALL | OPT_INCNOEXEC | OPT_SYM_OWNER | OPT_MULTI;
     parms.limited = -1;
 
     errmsg = ap_walk_config(conftree, &parms, s->lookup_defaults);
@@ -1648,12 +1766,15 @@ AP_DECLARE(void) ap_process_config_tree(server_rec *s,
                      parms.err_directive->filename);
         ap_log_perror(APLOG_MARK, APLOG_STARTUP, 0, p,
                      "%s", errmsg);
-        exit(1);
+        return HTTP_INTERNAL_SERVER_ERROR;
     }
+
+    return OK;
 }
 
 AP_CORE_DECLARE(int) ap_parse_htaccess(ap_conf_vector_t **result,
                                        request_rec *r, int override,
+                                       int override_opts,
                                        const char *d, const char *access_name)
 {
     ap_configfile_t *f = NULL;
@@ -1674,6 +1795,7 @@ AP_CORE_DECLARE(int) ap_parse_htaccess(ap_conf_vector_t **result,
 
     parms = default_parms;
     parms.override = override;
+    parms.override_opts = override_opts;
     parms.pool = r->pool;
     parms.temp_pool = r->pool;
     parms.server = r->server;
@@ -1730,6 +1852,7 @@ AP_CORE_DECLARE(int) ap_parse_htaccess(ap_conf_vector_t **result,
     new = apr_palloc(r->pool, sizeof(struct htaccess_result));
     new->dir = parms.path;
     new->override = override;
+    new->override_opts = override_opts;
     new->htaccess = dc;
 
     /* add to head of list */
@@ -1750,6 +1873,7 @@ AP_CORE_DECLARE(const char *) ap_init_virtual_host(apr_pool_t *p,
     s->process = main_server->process;
     s->server_admin = NULL;
     s->server_hostname = NULL;
+    s->server_scheme = NULL;
     s->error_fname = NULL;
     s->timeout = 0;
     s->keep_alive_timeout = 0;
@@ -1835,6 +1959,7 @@ static server_rec *init_server_config(process_rec *process, apr_pool_t *p)
     s->port = 0;
     s->server_admin = DEFAULT_ADMIN;
     s->server_hostname = NULL;
+    s->server_scheme = NULL;
     s->error_fname = DEFAULT_ERRORLOG;
     s->loglevel = DEFAULT_LOGLEVEL;
     s->limit_req_line = DEFAULT_LIMIT_REQUEST_LINE;
@@ -1867,15 +1992,20 @@ AP_DECLARE(server_rec*) ap_read_config(process_rec *process, apr_pool_t *ptemp,
                                        const char *filename,
                                        ap_directive_t **conftree)
 {
-    const char *confname;
+    const char *confname, *error;
     apr_pool_t *p = process->pconf;
     server_rec *s = init_server_config(process, p);
 
     init_config_globals(p);
 
     /* All server-wide config files now have the SAME syntax... */
-    process_command_config(s, ap_server_pre_read_config, conftree,
-                           p, ptemp);
+    error = process_command_config(s, ap_server_pre_read_config, conftree,
+                                   p, ptemp);
+    if (error) {
+        ap_log_error(APLOG_MARK, APLOG_STARTUP|APLOG_CRIT, 0, NULL, "%s: %s",
+                     ap_server_argv0, error);
+        return NULL;
+    }
 
     /* process_command_config may change the ServerRoot so
      * compute this config file name afterwards.
@@ -1886,13 +2016,24 @@ AP_DECLARE(server_rec*) ap_read_config(process_rec *process, apr_pool_t *ptemp,
         ap_log_error(APLOG_MARK, APLOG_STARTUP|APLOG_CRIT,
                      APR_EBADPATH, NULL, "Invalid config file path %s",
                      filename);
-        exit(1);
+        return NULL;
     }
 
-    ap_process_resource_config(s, confname, conftree, p, ptemp);
+    error = ap_process_resource_config(s, confname, conftree, p, ptemp);
+    if (error) {
+        ap_log_error(APLOG_MARK, APLOG_STARTUP|APLOG_CRIT, 0, NULL,
+                     "%s: %s", ap_server_argv0, error);
+        return NULL;
+    }
 
-    process_command_config(s, ap_server_post_read_config, conftree,
-                           p, ptemp);
+    error = process_command_config(s, ap_server_post_read_config, conftree,
+                                   p, ptemp);
+
+    if (error) {
+        ap_log_error(APLOG_MARK, APLOG_STARTUP|APLOG_CRIT, 0, NULL, "%s: %s",
+                     ap_server_argv0, error);
+        return NULL;
+    }
 
     return s;
 }
